@@ -9,6 +9,15 @@ public class StringToObjectConverter {
     public Object convertFromString(String pType, String pValue) {
         // TODO: Look for an external solution or support more types
         // At least use a map for lookup
+        if ("[null]".equals(pValue)) {
+            return null;
+        } else if ("\"\"".equals(pValue)) {
+            if (String.class.getName().equals(pType)) {
+                return "";
+            } else {
+                throw new IllegalArgumentException("Cannot convert empty string tag to type " + pType);
+            }
+        }
         if (String.class.getName().equals(pType)) {
             return pValue;
         } else if (Integer.class.getName().equals(pType) || "int".equals(pType)){
